
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include "defs.h"
#include "symtab.h"

SYMBOL_ENTRY symbol_table[SYMBOL_TABLE_LENGTH];
int first_empty = 0;


// Vraca indeks prvog sledeceg praznog elementa.
int get_next_empty_element(void) {
   if(first_empty < SYMBOL_TABLE_LENGTH)
      return first_empty++;
   else {
      printerror("Compiler error! Symbol table overflow!");
      exit(EXIT_FAILURE);
   }
}

// Vraca indeks poslednjeg zauzetog elementa.
int get_last_element(void) {
   return first_empty-1;
}

// Ubacuje simbol sa datom vrstom simbola i tipom simbola 
// i vraca indeks ubacenog elementa u tabeli simbola ili -1.
int insert_symbol(char *name, unsigned kind, unsigned type){
   int index = get_next_empty_element();
   symbol_table[index].name = name;
   symbol_table[index].kind = kind;
   symbol_table[index].type = type;
   return index;
}

// Proverava da li se simbol vec nalazi u tabeli simbola, 
// ako se ne nalazi ubacuje ga, ako se nalazi ispisuje gresku.
// Vraca indeks elementa u tabeli simbola.
int try_to_insert_id(char *name,unsigned kind,unsigned type){
   int index;
   if( (index = lookup_id(name, kind)) == -1 )
      index = insert_symbol(name, kind, type);
   else {
      printerror("redefinition of '%s'", name);
   }
   return index;
}

// Ubacuje konstantu u tabelu simbola (ako vec ne postoji).
int try_to_insert_constant(char *str, unsigned type) {
   int index;
   if( (index = lookup_constant(str, type)) == -1 ) {
      // provera opsega za konstante
      errno = 0;
      long int number = strtol(str, (char **) NULL, 10);
      // za 32 bitni int : INT_MAX = 2147483647 i 
      //                   INT_MIN = -2147483647-1 
      // micko kompajler je pravljen za 32-bitne platforme
      if( errno != 0 
          && (number == INT_MIN || number == INT_MAX) )   
         printerror("constant out of range");
      index = insert_symbol(str, CONSTANT, type);
   }
   return index;
}

// Vraca indeks pronadjenog simbola ili vraca -1.
int lookup_id(char *name, unsigned kind) {
   int i;
   for(i = first_empty - 1; i > FUNCTION_REG; i--) {
      if(strcmp(symbol_table[i].name, name) == 0 
         && symbol_table[i].kind & kind)
         return i;
   }
   return -1;
}

// Vraca indeks pronadjenog simbola (konstante) ili vraca -1.
int lookup_constant(char *name, unsigned type) {
   int i;
   for(i = first_empty - 1; i > FUNCTION_REG; i--) {
      if(strcmp(symbol_table[i].name, name) == 0 
         && symbol_table[i].type == type)
         return i;
   }
   return -1;
}


char *get_name(int index) {
   if(index > -1 && index < SYMBOL_TABLE_LENGTH)
      return symbol_table[index].name;
   return "?";
}

unsigned get_kind(int index) {
   if(index > -1 && index < SYMBOL_TABLE_LENGTH)
      return symbol_table[index].kind;
   return NO_KIND;
}

unsigned get_type(int index) {
   if(index > -1 && index < SYMBOL_TABLE_LENGTH)
      return symbol_table[index].type;
   return NO_TYPE;
}

void set_attribute(int index, int attribute) {
   if(index > -1 && index < SYMBOL_TABLE_LENGTH)
      symbol_table[index].attribute = attribute;
}

unsigned get_attribute(int index) {
   if(index > -1 && index < SYMBOL_TABLE_LENGTH)
      return symbol_table[index].attribute;
   return NO_ATTRIBUTE;
}

void set_register_type(int register_index, unsigned type) {
   if(register_index >= 0 
      && register_index <= FUNCTION_REG)
      symbol_table[register_index].type = type;
}

// Brise elemente tabele od zadatog indeksa do kraja tabele
void clear_symbols(unsigned begin_index) {
   int i;
   if(begin_index >= first_empty) {
      printerror("Compiler error! Wrong clear symbols argument");
      exit(EXIT_FAILURE);
   }
   for(i = begin_index; i < first_empty; i++) {
      if(symbol_table[i].name)
         free(symbol_table[i].name);
      symbol_table[i].name = 0;
      symbol_table[i].kind = NO_KIND;
      symbol_table[i].type = NO_TYPE;
      symbol_table[i].attribute = NO_ATTRIBUTE;
   }
   first_empty = begin_index;
}

// Brise sve elemente tabele simbola.
void clear_symtab(void) {
   first_empty = SYMBOL_TABLE_LENGTH - 1;
   clear_symbols(0);
}
   
// Ispisuje sve elemente tabele simbola.
void print_symtab(void) {
   static const char *symbol_kinds[] = { "NONE", "REGISTER", 
      "CONSTANT", "FUNCTION", "LOCALVAR" };
   int i,j;
   printf("\n\nSYMBOL TABLE\n");
   printf("\n       name           kind   type  attr");
   printf("\n-- ---------------- -------- ----  ----");
   for(i = 0; i < first_empty; i++) {
      printf("\n%2d %-16s %-8s %4d %4d ", i, 
      symbol_table[i].name, 
      symbol_kinds[(int)(logarithm2(symbol_table[i].kind))], 
      symbol_table[i].type, symbol_table[i].attribute);
   }
   printf("\n\n");
}

unsigned logarithm2(unsigned value) {
   unsigned mask = 1;
   int i = 0;
   for(i = 0; i < 32; i++) {
      if(value & mask)
         return i;
      mask <<= 1;
   }
   return 0; // ovo ne bi smelO; indeksiraj string "NONE"
}

// Inicijalizacija tabele simbola.
void init_symtab(void) {
   clear_symtab();

   int i = 0;
   char s[4];
   for(i = 0; i < 14; i++) {
      sprintf(s, "%%%d", i);
      insert_symbol(strdup(s), REGISTER, NO_TYPE);
   }
}

